#include <msp/strings/format.h>
#include "getopt.h"

using namespace std;

namespace Msp {

GetOpt::GetOpt():
	help(false)
{
	add_option("help", help, NO_ARG).set_help("Displays this help");
}

GetOpt::~GetOpt()
{
	for(list<OptBase *>::iterator i=opts.begin(); i!=opts.end(); ++i)
		delete *i;
}

GetOpt::OptBase &GetOpt::add_option(OptBase *opt)
{
	for(list<OptBase *>::iterator i=opts.begin(); i!=opts.end(); )
	{
		if((opt->get_short()!=0 && (*i)->get_short()==opt->get_short()) || (*i)->get_long()==opt->get_long())
		{
			delete *i;
			opts.erase(i++);
		}
		else
			++i;
	}

	opts.push_back(opt);
	return *opts.back();
}

GetOpt::OptBase &GetOpt::get_option(char s)
{
	for(list<OptBase *>::iterator i=opts.begin(); i!=opts.end(); ++i)
		if((*i)->get_short()==s)
			return **i;
	throw usage_error(string("Unknown option -")+s);
}

GetOpt::OptBase &GetOpt::get_option(const string &l)
{
	for(list<OptBase *>::iterator i=opts.begin(); i!=opts.end(); ++i)
		if((*i)->get_long()==l)
			return **i;
	throw usage_error(string("Unknown option --")+l);
}

void GetOpt::operator()(unsigned argc, const char *const *argv)
{
	try
	{
		unsigned i = 1;
		for(; i<argc;)
		{
			if(argv[i][0]=='-')
			{
				if(argv[i][1]=='-')
				{
					if(!argv[i][2])
						break;

					i += process_long(argv+i);
				}
				else
					i += process_short(argv+i);
			}
			else
				args.push_back(argv[i++]);
		}
		
		for(; i<argc; ++i)
			args.push_back(argv[i]);
	}
	catch(const usage_error &e)
	{
		throw usage_error(e.what(), "Usage: "+generate_usage(argv[0]));
	}

	if(help)
		throw usage_error(string("Help for ")+argv[0]+":", generate_help());
}

unsigned GetOpt::process_long(const char *const *argp)
{
	// Skip the --
	const char *arg = argp[0]+2;

	// See if the argument contains an =
	unsigned equals = 0;
	for(; arg[equals] && arg[equals]!='='; ++equals) ;
	
	OptBase &opt = get_option(string(arg, equals));
	
	if(arg[equals])
		// Process the part after the = as option argument
		opt.process(arg+equals+1);
	else if(opt.get_arg_type()==REQUIRED_ARG)
	{
		if(!argp[1])
			throw usage_error("--"+string(arg)+" requires an argument");

		// Process the next argument as option argument
		opt.process(argp[1]);
		return 2;
	}
	else
		opt.process();
	
	return 1;
}

unsigned GetOpt::process_short(const char *const *argp)
{
	// Skip the -
	const char *arg = argp[0]+1;

	// Loop through all characters in the argument
	for(; *arg; ++arg)
	{
		OptBase &opt = get_option(*arg);

		if(arg[1] && opt.get_arg_type()!=NO_ARG)
		{
			// Need an option argument and we have characters left - use them
			opt.process(arg+1);
			return 1;
		}
		else if(opt.get_arg_type()==REQUIRED_ARG)
		{
			if(!argp[1])
				throw usage_error("-"+string(1, *arg)+" requires an argument");
			
			// Use the next argument as option argument
			opt.process(argp[1]);
			return 2;
		}
		else
			opt.process();
	}

	return 1;
}

string GetOpt::generate_usage(const string &argv0) const
{
	string result = argv0;
	for(list<OptBase *>::const_iterator i=opts.begin(); i!=opts.end(); ++i)
	{
		result += " [";
		if((*i)->get_short())
		{
			result += format("-%c", (*i)->get_short());
			if(!(*i)->get_long().empty())
				result += '|';
			else if((*i)->get_arg_type()==OPTIONAL_ARG)
				result += format("[%s]", (*i)->get_metavar());
			else if((*i)->get_arg_type()==REQUIRED_ARG)
				result += format(" %s", (*i)->get_metavar());
		}
		if(!(*i)->get_long().empty())
		{
			result += format("--%s", (*i)->get_long());

			if((*i)->get_arg_type()==OPTIONAL_ARG)
				result += format("[=%s]", (*i)->get_metavar());
			else if((*i)->get_arg_type()==REQUIRED_ARG)
				result += format("=%s", (*i)->get_metavar());
		}
		result += ']';
	}

	return result;
}

string GetOpt::generate_help() const
{
	bool any_short = false;
	for(list<OptBase *>::const_iterator i=opts.begin(); (!any_short && i!=opts.end()); ++i)
		any_short = (*i)->get_short();

	string::size_type maxw = 0;
	list<string> switches;
	for(list<OptBase *>::const_iterator i=opts.begin(); i!=opts.end(); ++i)
	{
		string swtch;
		if((*i)->get_short())
		{
			swtch += format("-%c", (*i)->get_short());
			if(!(*i)->get_long().empty())
				swtch += ", ";
			else if((*i)->get_arg_type()==OPTIONAL_ARG)
				swtch += format("[%s]", (*i)->get_metavar());
			else if((*i)->get_arg_type()==REQUIRED_ARG)
				swtch += format(" %s", (*i)->get_metavar());
		}
		else if(any_short)
			swtch += "    ";
		if(!(*i)->get_long().empty())
		{
			swtch += format("--%s", (*i)->get_long());

			if((*i)->get_arg_type()==OPTIONAL_ARG)
				swtch += format("[=%s]", (*i)->get_metavar());
			else if((*i)->get_arg_type()==REQUIRED_ARG)
				swtch += format("=%s", (*i)->get_metavar());
		}
		switches.push_back(swtch);
		maxw = max(maxw, swtch.size());
	}

	string result;
	list<string>::const_iterator j = switches.begin();
	for(list<OptBase *>::const_iterator i=opts.begin(); i!=opts.end(); ++i, ++j)
		result += format("  %s%s%s\n", *j, string(maxw+2-j->size(), ' '), (*i)->get_help());
	
	return result;
}


GetOpt::OptBase::OptBase(char s, const std::string &l, ArgType a):
	shrt(s),
	lng(l),
	arg_type(a),
	seen_count(0),
	metavar("ARG")
{
	if(lng.empty())
		throw invalid_argument("empty long option name");
}

GetOpt::OptBase &GetOpt::OptBase::set_help(const string &h)
{
	help = h;
	return *this;
}

GetOpt::OptBase &GetOpt::OptBase::set_help(const string &h, const string &m)
{
	help = h;
	metavar = m;
	return *this;
}

void GetOpt::OptBase::process()
{
	if(arg_type==REQUIRED_ARG)
		throw usage_error("--"+lng+" requires an argument");
	++seen_count;

	store();
}

void GetOpt::OptBase::process(const string &arg)
{
	if(arg_type==NO_ARG)
		throw usage_error("--"+lng+" takes no argument");
	++seen_count;

	store(arg);
}

} // namespace Msp
