#include <cstdlib>
#include <cerrno>
#include <dirent.h>
#ifdef WIN32
#include <shlobj.h>
#else
#include <unistd.h>
#include <sys/stat.h>
#endif
#include <msp/core/systemerror.h>
#include <msp/strings/regex.h>
#include <msp/strings/utils.h>
#include "dir.h"
#include "path.h"
#include "stat.h"
#include "utils.h"

using namespace std;

namespace Msp {
namespace FS {

namespace
{

/** Helper function to determine the location of the program's executable.
Caches the last result to cut down filesystem access with repeated calls. */
const Path &get_bin_dir(const string &argv0)
{
	static string last_argv0;
	static Path bin_dir;

	if(!(argv0==last_argv0))
	{
		Path exe;
		if(argv0.find('/')==string::npos)
		{
			const char *path = getenv("PATH");
			vector<string> dirs = split(path, ':');
			for(vector<string>::const_iterator i=dirs.begin(); i!=dirs.end(); ++i)
				if(exists(Path(*i)/argv0))
				{
					exe = realpath(Path(*i)/argv0);
					break;
				}
		}
		else
			exe = realpath(argv0);

		last_argv0 = argv0;
		bin_dir = dirname(exe);
	}

	return bin_dir;
}

}


not_a_directory::not_a_directory(const Path &p):
	runtime_error(p.str())
{ }


void mkdir(const Path &path, int mode)
{
#ifdef WIN32
	(void)mode;
	if(!CreateDirectory(path.str().c_str(), NULL))
		throw system_error("CreateDirectory");
#else
	if(::mkdir(path.str().c_str(), mode)==-1)
		throw system_error("mkdir");
#endif
}

void mkpath(const Path &path, int mode)
{
	Path p;
	for(Path::Iterator i=path.begin(); i!=path.end(); ++i)
	{
		p /= *i;
#ifdef WIN32
		if(p.size()==1 && p.is_absolute())
			continue;
#endif
		if(exists(p))
		{
			if(!is_dir(p))
				throw not_a_directory(p);
			continue;
		}
		else
			mkdir(p, mode);
	}
}

void rmdir(const Path &path)
{
#ifdef WIN32
	if(!RemoveDirectory(path.str().c_str()))
		throw system_error("RemoveDirectory");
#else
	if(::rmdir(path.str().c_str())==-1)
		throw system_error("rmdir");
#endif
}

void rmpath(const Path &path)
{
	list<string> files = list_files(path);
	for(list<string>::iterator i=files.begin(); i!=files.end(); ++i)
	{
		Path p = path / *i;
		if(is_dir(p))
			rmpath(p);
		else
			unlink(p);
	}

	rmdir(path);
}

list<string> list_files(const Path &path)
{
	return list_filtered(path, string());
}

list<string> list_filtered(const Path &path, const string &filter)
{
	Regex r_filter(filter);

	list<string> result;
	DIR *dir = opendir(path.str().c_str());
	if(dir)
	{
		while(dirent *de = readdir(dir))
		{
			const char *fn = de->d_name;
			if(fn[0]=='.' && (fn[1]==0 || (fn[1]=='.' && fn[2]==0))) 
				continue;
			if(r_filter.match(fn))
				result.push_back(fn);
		}
		closedir(dir);
	}

	return result;
}

Path getcwd()
{
	char buf[1024];
	return ::getcwd(buf, sizeof(buf));
}

Path get_home_dir()
{
#ifdef WIN32
	char home[MAX_PATH];
	if(SHGetFolderPath(0, CSIDL_PERSONAL, 0, 0, home)==S_OK)
		return home;
#else
	const char *home = getenv("HOME");
	if(home)
		return home;
#endif
	return ".";
}

Path get_user_data_dir(const string &appname)
{
#ifdef WIN32
	char datadir[MAX_PATH];
	if(SHGetFolderPath(0, CSIDL_LOCAL_APPDATA, 0, 0, datadir)==S_OK)
		return Path(datadir)/appname;
	return ".";
#else
	return get_home_dir()/("."+appname);
#endif
}

Path get_sys_conf_dir(const string &argv0)
{
	Path dir = get_bin_dir(argv0);

	if(dir[-1]=="bin" || dir[-1]=="sbin")
	{
		dir /= "..";
		if(dir[-1]=="usr")
			dir /= "..";
		return dir/"etc";
	}
	else
		return dir;
}

Path get_sys_data_dir(const string &argv0, const string &appname)
{
	Path dir = get_bin_dir(argv0);

	if(dir[-1]=="bin" || dir[-1]=="sbin")
		return dir/".."/"share"/appname;
	else
		return dir;
}

Path get_sys_lib_dir(const string &argv0, const string &appname)
{
	Path dir = get_bin_dir(argv0);

	if(dir[-1]=="bin" || dir[-1]=="sbin")
		return dir/".."/"lib"/appname;
	else
		return dir;
}

void chdir(const Path &path)
{
	if(::chdir(path.str().c_str())==-1)
		throw system_error("chdir");
}

} // namespace FS
} // namespace Msp
